/**
 * @file
 * @brief Item related wizard functions.
**/

#pragma once

void wizard_create_spec_object();
void wizard_create_spec_object_by_name();
void wizard_tweak_object();
void wizard_make_object_randart();
void wizard_value_item();
void wizard_uncurse_item();
void wizard_create_all_artefacts();
void wizard_identify_pack();
void wizard_unidentify_pack();
void wizard_draw_card();
void wizard_list_items();
void wizard_identify_all_items();
void wizard_unidentify_all_items();
void debug_item_statistics();
void wizard_recharge_evokers();
