#pragma once

enum god_type
{
    GOD_NO_GOD = 0,
    GOD_ZIN,
    GOD_SHINING_ONE,
    GOD_KIKUBAAQUDGHA,
    GOD_YREDELEMNUL,
    GOD_XOM,
    GOD_VEHUMET,
    GOD_OKAWARU,
    GOD_MAKHLEB,
    GOD_SIF_MUNA,
    GOD_TROG,
    GOD_NEMELEX_XOBEH,
    GOD_ELYVILON,
    GOD_LUGONU,
    GOD_BEOGH,
    GOD_JIYVA,
    GOD_FEDHAS,
    GOD_CHEIBRIADOS,
    GOD_ASHENZARI,
    GOD_DITHMENOS,
    GOD_GOZAG,
    GOD_QAZLAL,
    GOD_RU,
    GOD_PAKELLAS,
    GOD_USKAYAW,
    GOD_HEPLIAKLQANA,
    GOD_WU_JIAN,
    NUM_GODS,                          // always after last god

    GOD_RANDOM = 100,
    GOD_NAMELESS,                      // for monsters with non-player gods
    GOD_ECUMENICAL,                    // Temporary
};
