/**
 * @file
 * @brief Just because typedefs can't be forward declared. Argh.
**/

#pragma once

#include "defines.h"

typedef int travel_distance_col[GYM];
typedef travel_distance_col travel_distance_grid_t[GXM];
