Play now:
{% set rcs = set() %}
{% for id in games %}
{% if games[id]["rcfile_path"] not in rcs %}
{% if "separator" in games[id] %}
{% raw games[id]["separator"] %}
{% else %}<br>{% end %}
<a href="#play-{{ id }}">{{ games[id]["name"] }}</a>
{% set rcs.add(games[id]["rcfile_path"]) %}
<a href="javascript:"
   class="edit_rc_link"
   data-game_id="{{ id }}">(edit rc)</a>
{% else %}
{% if "separator" in games[id] %}
{% raw games[id]["separator"] %}
{% else %} | {% end %}
<a href="#play-{{ id }}">{{ games[id]["name"] }}</a>
{% end %}
{% end %}
