#pragma once

enum torment_source_type
{
    TORMENT_LURKING_HORROR= -1,
    TORMENT_CARDS         = -2,   // Symbol of torment
    TORMENT_SCEPTRE       = -3,   // The Sceptre of Torment
    TORMENT_SCROLL        = -4,
    TORMENT_SPELL         = -5,   // SPELL_SYMBOL_OF_TORMENT
    TORMENT_XOM           = -6,   // Xom effect
    TORMENT_KIKUBAAQUDGHA = -7,   // Kikubaaqudgha effect
    TORMENT_MISCAST       = -8,
    TORMENT_AGONY         = -9,   // SPELL_AGONY
};
