body {
    background: black;
    color: white;
    font: 16px monospace;
}

#login {
    text-align: right;
    float: right;
}

#crt {
    display: inline-block;
}

#lobby_body {
    text-align: center;
}

#prompt {
    top: 50%;
    left: 50%;
    position: absolute;
    z-index: 4000;
}

#prompt > div {
    background-color: black;
    width: 34em;
    margin-left: -17em;
    margin-top: -10ex;
    padding: 1em 2ex;
    border: 1px solid white;
}

#prompt #login {
    float: none;
    text-align: left;
    margin: 2ex 0;
}

#prompt_game {
    font-weight: bold;
}

#player_list {
    width: 100%;
    text-align: left;
    line-height: 1.1em;
}

#player_list th.header {
    cursor: pointer;
    padding-left: 1px;
    padding-right: 17px;
}
#player_list th.headerSortDown {
    background-image: url("asc.gif");
    background-repeat: no-repeat;
    background-position: center left;
    padding-left: 17px;
    padding-right: 1px;
}
#player_list th.headerSortUp {
    background-image: url("desc.gif");
    background-repeat: no-repeat;
    background-position: center left;
    padding-left: 17px;
    padding-right: 1px;
}

#player_list th.xl.header {
    width: 2em;
    text-align: right;
    padding-left: 17px;
    padding-right: 5px;
}
#player_list th.xl.headerSortDown,
#player_list th.xl.headerSortUp {
    padding-right: 5px;
}
#player_list td.xl {
    text-align: right;
    padding-right: 5px;
}
#player_list th.char {
    width: 3em;
}
#player_list th.place {
    width: 3em;
}

#player_list th.idle_time {
    width: 3em;
}
#player_list th.spectator_count {
    width: 1em;
}

#player_list th.milestone_col {
    width: 20em;
}

#player_list tr:hover {
    background-color: #101010;
}

#player_list .milestone_container {
    overflow-y: hidden;
    height: 1.1em;
}
#player_list .milestone {
    position: relative;
}
#player_list .new_milestone {
    position: relative;
}

.floating_dialog {
    position: absolute;
    left: 50px;
    top: 50px;

    border: 1px solid lightgray;
    padding: 10px;
    background-color: black;

    z-index: 3000;
}

.floating_dialog .title {
    text-align: center;
}

#register label {
    width: 200px;
    float: left;
}

#reg_cancel {
    float: left;
}

#reg_submit {
    float: right;
}

.text {
    background-color: black;
    color: white;
    border: 1px solid white;
}
.button {
    background-color: #DDDDDD;
    border: 1px solid white;
}

#exit_game_dump {
    text-align: right;
}

#chat {
    background: rgba(0, 0, 0, 0.85);
    width: 400px;
    position: fixed;
    bottom: 20px;
    right: 20px;

    color: #A2A2A2;

    font-size: small;

    padding: 5px;
    border: 1px solid #626262;

    z-index: 3000;
}

#chat a {
    color: inherit;
    font-weight: inherit;
}

#chat_caption {
    font-weight: normal;
    text-decoration: none;
}

#chat_caption:hover {
    background-color: #111111;
}

#chat_login_text {
    border-top: 1px dashed #626262;
}

.chat_automated {
    color: white;
    font-style: italic;
}

#message_count {
    float: right;
}
#message_count.has_new {
    font-weight: bolder;
    color: green;
}

#chat_history_container {
    width: 100%;
    max-height: 200px;
    overflow-y: auto;
    overflow-x: hidden;
    text-overflow: ellipsis;
    word-wrap: break-word;
    margin-bottom: 3px;
}

#chat_history {
    width: 100%;
}

.chat_sender {
    font-weight: bold;
}

#spectator_list .player {
    font-weight: bold;
}

#spectator_list .watcher {
    font-weight: normal;
}

#chat_input {
    width: 394px;
    font: small monospace;
}

a:link, a:active, a:visited {
    color: white;
    font-weight: bold;
}

a:hover {
    color: lightgrey;
    font-weight: bold;
}

#loader {
    display: table;
    text-align: center;
    width: 100%;
    margin: 0px;
    padding: 0px;
    position: fixed;
    top: 0;
    left: 0;
}
#loader_center {
    display: table-cell;
    vertical-align: middle;
}

#overlay {
    display: none;
    position: fixed;
    left: 0;
    top: 0;
    width: 100%;
    height: 100%;
    background-color: black;
    opacity: 0.7;
    z-index: 2000;
}

/* These are just here for compatibility with older versions */
.fg0 { color: rgb(  0,   0,   0); }
.fg1 { color: rgb(  0,  82, 255); }
.fg2 { color: rgb(100, 185,  70); }
.fg3 { color: rgb(  0, 180, 180); }
.fg4 { color: rgb(255,  48,   0); }
.fg5 { color: rgb(238,  92, 238); }
.fg6 { color: rgb(165,  91,   0); }
.fg7 { color: rgb(162, 162, 162); }
.fg8 { color: rgb( 82,  82,  82); }
.fg9 { color: rgb( 82, 102, 255); }
.fg10 { color: rgb( 82, 255,  82); }
.fg11 { color: rgb( 82, 255, 255); }
.fg12 { color: rgb(255,  82,  82); }
.fg13 { color: rgb(255,  82, 255); }
.fg14 { color: rgb(255, 255,  82); }
.fg15 { color: rgb(255, 255, 255); }

.bg0 { background-color: rgb(  0,   0,   0); }
.bg1 { background-color: rgb(  0,  82, 255); }
.bg2 { background-color: rgb(100, 185,  70); }
.bg3 { background-color: rgb(  0, 180, 180); }
.bg4 { background-color: rgb(255,  48,   0); }
.bg5 { background-color: rgb(238,  92, 238); }
.bg6 { background-color: rgb(165,  91,   0); }
.bg7 { background-color: rgb(162, 162, 162); }
.bg8 { background-color: rgb( 82,  82,  82); }
.bg9 { background-color: rgb( 82, 102, 255); }
.bg10 { background-color: rgb( 82, 255,  82); }
.bg11 { background-color: rgb( 82, 255, 255); }
.bg12 { background-color: rgb(255,  82,  82); }
.bg13 { background-color: rgb(255,  82, 255); }
.bg14 { background-color: rgb(255, 255,  82); }
.bg15 { background-color: rgb(255, 255, 255); }

#stats {
    display: inline-block;
    vertical-align: top;
    position: absolute;
}

#minimap, #minimap_overlay {
    position: absolute;
}
